<!DOCTYPE html>

{# make the macros shipped with hansken.py available on a 'hansken' namespace #}
{% import 'hansken/macros.html' as hansken %}

<html lang="en">

<head>
    <meta charset="UTF-8">
    <title>{% block title %}{{ title }}{% endblock %}</title>

    {# provide user-overridable default styles, making it possible to override styles altogether #}
    {% block styles %}
        <style type="text/css">
            {% include 'hansken/default.css' %}
        </style>
        <style type="text/css" media="print">
            {% include 'hansken/print.css' %}
        </style>
        {% block extra_styles %}{% endblock %}
    {% endblock %}

    {# provide user-overridable default scripts, making it possible to override scripts altogether #}
    {% block scripts %}
        {% block extra_scripts %}{% endblock %}
    {% endblock %}
</head>

<body>

{# split content in three parts, with hansken.py-shipped templates only using 'content' #}
{% block preamble %}{{ preamble }}{% endblock %}

{% block content %}{{ content }}{% endblock %}

{% block postamble %}{{ postamble }}{% endblock %}

</body>

</html>
